import logging
import os
import pexpect
import re
import requests
from io import BytesIO
import tarfile
from functools import lru_cache
from io import StringIO
from html.parser import HTMLParser
from ._version import __version__, __version_tuple__


__all__ = [
    "__version__",
    "__version_tuple__",
    "command",
    "download",
    "get_mirror",
    "get_size",
    "is_current",
    "URL",
    "OLDURL",
]


log = logging.getLogger(__name__)

URL = 'https://mirror.ctan.org/systems/texlive/tlnet/'
OLDURL = 'https://ftp.tu-chemnitz.de/pub/tug/historic/systems/texlive/{v}/tlnet-final/'


class GetText(HTMLParser):
    """Only extract text of html page"""

    def __init__(self):
        super().__init__()
        self._text = StringIO()

    def handle_data(self, d):
        self._text.write(d)

    @property
    def text(self):
        return self._text.getvalue()


@lru_cache
def is_current(version):
    r = requests.get('https://tug.org/texlive/')
    r.raise_for_status()

    parser = GetText()
    parser.feed(r.text)

    m = re.search(r'Current release: TeX Live ([0-9]{4})', parser.text)
    if not m:
        raise ValueError('Could not determine current TeX Live version')

    current_version = int(m.groups()[0])
    log.debug('Current version of TeX Live is {}'.format(current_version))
    return current_version == version


def get_mirror():
    """Get a CTAN mirror"""
    r = requests.get(URL, allow_redirects=False)
    r.raise_for_status()
    return r.headers["Location"]


def download(version=None, outdir='.', url=None):
    os.makedirs(outdir, exist_ok=True)

    if version is None or is_current(version):
        url = url or URL
        url = url.rstrip("/") + '/install-tl-unx.tar.gz'
    else:
        url = OLDURL.format(v=version) + 'install-tl-unx.tar.gz'

    log.debug('Downloading from {}'.format(url))

    ret = requests.get(url)
    ret.raise_for_status()

    tar = tarfile.open(fileobj=BytesIO(ret.content), mode='r:gz')
    tar.extractall(outdir)


def command(process, pattern, send, **kwargs):
    process.expect(pattern, **kwargs)
    process.sendline(send)


def get_size(process):
    timeout = process.timeout
    process.timeout = 1
    lines = ''
    try:
        while True:
            lines += process.readline().decode()
    except pexpect.TIMEOUT:
        pass

    size = re.findall(r'disk space required: ([0-9]+ MB)', lines)[-1]
    process.timeout = timeout
    return size



from argparse import ArgumentParser
from . import __version__

parser = ArgumentParser(prog='install_texlive')

parser.add_argument('-V', '--version', action='version', version=__version__)

parser.add_argument(
    '-v', '--verbose', action='store_true', help='Create more verbose output'
)

parser.add_argument(
    '-t', '--texlive-version', dest='version', type=int,
    help='TeX Live version to install',
)

parser.add_argument(
    '--install-tl',
    help='Path to the install-tl script. If not given, TeX Live will be downloaded.',
)

parser.add_argument(
    '-k', '--keep-config', action='store_true',
    help='If an existing installation is found, keep its config',
)

parser.add_argument(
    '-p', '--prefix',
    help='Installation prefix, equivalent to setting TEXLIVE_INSTALL_PREFIX'
)

parser.add_argument(
    '-c', '--collections',
    help=(
        'The TeX Live package collections to install.'
        ' E.g. -a to deselect all and only install the absolute basic TeX packages'
    )
)

parser.add_argument(
    '-s', '--scheme', choices=set('abcdefghijk'),
    help='The TeX Live scheme to install. Default is "full"'
)

parser.add_argument(
    '-u', '--update', action='store_true',
    help='Update TeX Live after installation is finished'
)

parser.add_argument(
    '-i', '--install',
    help='Additional packages to install after the main Installation has finished as comma separated list'
)

parser.add_argument(
    '-f', '--package-file',
    help='A file with one package per line to be install after the main installation has finished'
)

parser.add_argument(
    '--source', action='store_true',
    help='Install the source tree',
)

parser.add_argument(
    '--docs', action='store_true',
    help='Install the docs tree',
)

parser.add_argument(
    "-r", "--repository",
    help="If given, set ctan mirror to this URL.",
)



import pexpect
from glob import glob
import os
import sys
import logging
import tempfile
import re
import subprocess as sp

from . import command, download, OLDURL, get_size, is_current, get_mirror
from .parser import parser


logging.basicConfig(level=logging.INFO)
log = logging.getLogger('install_texlive')

timeout = 30


def main():
    args = parser.parse_args()

    if args.verbose:
        log.level = logging.DEBUG

    if args.prefix:
        args.prefix = os.path.abspath(args.prefix)
        os.environ['TEXLIVE_INSTALL_PREFIX'] = args.prefix
        os.makedirs(args.prefix, exist_ok=True)

    os.environ["TEXLIVE_INSTALL_ENV_NOCHECK"] = "1"
    log.info('Installing texlive to {}'.format(args.prefix or '/usr/local/texlive'))

    if args.repository is None:
        if args.version is None or is_current(args.version):
            args.repository = get_mirror()
        else:
            args.repository = OLDURL.format(v=args.version)

    if args.install_tl:
        install_script = args.install_tl
        cmd = install_script
    else:
        log.info("Using repository: %s", args.repository)
        directory = os.path.join(
            tempfile.gettempdir(), 'texlive-{}'.format(args.version or 'current')
        )
        download(version=args.version, outdir=directory, url=args.repository)
        install_script = glob(os.path.join(directory, 'install-tl-*/install-tl'))[-1]
        cmd = install_script + ' --repository={}'.format(args.repository) 

    log.info(cmd)

    tl = pexpect.spawn(cmd, timeout=timeout)

    try:
        command(tl, 'installation.profile', 'N', timeout=5)
    except pexpect.TIMEOUT:
        pass

    try:
        command(tl, 'Import settings', 'y' if args.keep_config else 'n', timeout=5)
    except pexpect.TIMEOUT:
        log.info('No previous installation found')

    try:
        if args.scheme:
            command(tl, 'Enter command:', 'S', timeout=timeout)
            command(tl, 'Enter letter', args.scheme, timeout=timeout)
            command(tl, 'Enter letter', 'R', timeout=timeout)

        if args.collections:
            command(tl, 'Enter command:', 'C', timeout=timeout)
            command(tl, 'Enter letter', args.collections, timeout=timeout)
            command(tl, 'Enter letter', 'R', timeout=timeout)

        if not args.docs:
            command(tl, 'Enter command:', 'O', timeout=timeout)
            command(tl, 'Enter command:', 'D', timeout=timeout)
            command(tl, 'Enter command:', 'R', timeout=timeout)

        if not args.source:
            command(tl, 'Enter command:', 'O', timeout=timeout)
            command(tl, 'Enter command:', 'S', timeout=timeout)
            command(tl, 'Enter command:', 'R', timeout=timeout)

        log.info('Installation size will be {}'.format(get_size(tl)))
        log.info('Starting installation')
        command(tl, 'Enter command:', 'I', timeout=timeout)
    except pexpect.TIMEOUT:
        log.error('Something went wrong')
        sys.exit(1)

    lines = ''
    try:
        while not tl.terminated:
            line = tl.readline().decode()
            log.debug(line.strip())
            lines += line
    except pexpect.EOF:
        log.error('EOF')

    tl.close()
    if tl.exitstatus == 0:
        log.info('Installation installed succesfully')
    else:
        log.error('Installation did not finish succesfully')
        sys.exit(tl.exitstatus)

    bindir = re.findall(r'Most importantly, add\s+(.*)\s+to your PATH', lines)[0].strip()
    version = re.findall(r'20[0-9][0-9]', bindir)[0]
    env = os.environ
    env['PATH'] = os.path.abspath(bindir) + ':' + env['PATH']

    repo = args.repository
    if args.version is not None and not is_current(args.version):
        repo = OLDURL.format(v=version)

    if repo is not None:
        log.info("Setting repository to %s", repo)
        sp.run(
            ['tlmgr', 'option', 'repository', repo],
            env=env,
            check=True,
        )

    if args.update:
        log.info('Start updating')
        sp.run(
            ['tlmgr', 'update', '--self', '--all', '--reinstall-forcibly-removed'],
            env=env,
            check=True,
        )
        log.info('Finished')

    additional_packages = []
    if args.install:
        additional_packages.extend(args.install.split(','))

    if args.package_file:
        with open(args.package_file, 'r') as f:
            additional_packages.extend(f.read().splitlines())

    if additional_packages:
        log.info('Start installing additional packages')
        # tlmgr must always be up to date to install packages
        sp.run(['tlmgr', 'update', '--self'], env=env, check=True)
        sp.run(['tlmgr', 'install', *additional_packages], env=env, check=True)
        log.info('Finished')


if __name__ == '__main__':
    main()
